﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                
//                                           
using System.Collections.Generic;
using UnityEngine;

namespace PuppetMaster
{
    public enum AttackState
    {
        ready,
        windup,
        hit,
        final,
    }


    public class Attack
    {
        public Puppet Puppet = null;
        public AttackState state      = AttackState.ready;
        public AttackIds CurrentAttack;
        public MoveTypes CurrentMove;
        public Thing thing;

        public PersonBehaviour Enemy;
        public Vector2 EnemyTarget;
        public Vector2 TargetPos;
        public float TargetAngle;
        public Dictionary<string, Rigidbody2D> XRB2 = new Dictionary<string, Rigidbody2D>();

        public float floatemp1;
        public float floatemp2;
        public float power      = 0f;
        public float facing     = 1f;
        public int frame        = 0;
        public float currentVelocity;

        public MoveSet combatPose = new MoveSet("UpperArmFront:-11.68893:10,LowerArmFront:-96.55749:10");
        public List<MoveSet> AttackMoves = new List<MoveSet>();

        public enum AttackIds
        {
            club,
            thrust,
            thrustx,
            dislodgeKick,
            dislodgeBack,
            dislodgeIchi,
        }

        public enum MoveTypes
        {
            stab,
            slash,
        }

        public Vector2 Vectemp1;
        public Vector2 Vectemp2;




        //
        // ─── TARGET ENEMY ────────────────────────────────────────────────────────────────
        //
        public void TargetEnemy()
        {
            PersonBehaviour TheChosen = (PersonBehaviour)null;
            PersonBehaviour[] people  = GameObject.FindObjectsOfType<PersonBehaviour>();
            floatemp1                 = float.MaxValue;

            bool lastKnockedOut       = false;

            foreach (PersonBehaviour person in people)
            {
                if (!person.isActiveAndEnabled || !person.IsAlive() || person == Puppet.PBO) continue;

                Vectemp1 = person.transform.GetComponentInChildren<Rigidbody2D>().position;

                if (Puppet.FacingLeft  && Vectemp1.x > Puppet.RB2["Head"].position.x) continue;
                if (!Puppet.FacingLeft && Vectemp1.x < Puppet.RB2["Head"].position.x) continue;

                floatemp2 = (thing.R.position - Vectemp1).sqrMagnitude;

                if (floatemp2 < floatemp1)
                {
                    floatemp1 = floatemp2;
                    TheChosen = person;
                    if (person.Consciousness < 1 || person.ShockLevel > 0.3f)
                    {
                        lastKnockedOut = true;
                    }
                } else if(lastKnockedOut)
                {
                    if (person.Consciousness >= 1 && person.ShockLevel < 0.3f)
                    {
                        if (floatemp2 - floatemp1 < 1f)
                        {
                            lastKnockedOut = false;
                            floatemp1 = floatemp2;
                            TheChosen = person;
                        }
                    }
                }
            }

            //if (TheChosen == null) Enemy = TheChosen;

            //if (TheChosen == Enemy) return;

            Enemy = TheChosen;

            if (TheChosen == null) {

                EnemyTarget = Puppet.RB2["UpperBody"].position + (Vector2.right * facing * 1.5f);
                return;

            }

            string[] ValidTargets = new string[]
            {
                "Head",
                "UpperBody",
                "MiddleBody",
                "LowerBody",
                "UpperArm",
                "UpperArmFront",
                "LowerArm",
                "LowerArmFront",
            };

            string randomTarget = ValidTargets[UnityEngine.Random.Range(0,ValidTargets.Length - 1)];  
            

            XRB2.Clear();

            Rigidbody2D[] RBs = Enemy.transform.GetComponentsInChildren<Rigidbody2D>();


            foreach (Rigidbody2D rb in RBs)
            {
                XRB2.Add(rb.name, rb);
            }

            EnemyTarget = XRB2[ randomTarget ].position;

        }

        //
        // ─── ATTACK INIT ────────────────────────────────────────────────────────────────
        //
        public void Init(AttackIds attackId)
        {

            if (state != AttackState.ready) return;

            if (!Puppet.PBO.IsTouchingFloor) return;

            thing                   = Puppet.HoldingF;
            thing.P.ForceContinuous = true;
            thing.AttackDamage      = 0f;
            frame                   = 0;
            power                   = 0;
            state                   = AttackState.windup;
            facing                  = Puppet.FacingLeft ? -1 : 1;
            state                   = AttackState.windup;
            Puppet.IsWalking        = false;

            thing.AttackedList.Clear();

            AttackMoves.Clear();

            CurrentAttack = attackId;

            if (!Puppet.IsCrouching) Puppet.PBO.OverridePoseIndex = -1;

            if (attackId == AttackIds.club)
            {
                TargetEnemy();

                AttackMoves.Add(new MoveSet(@"
                        UpperArmFront:169.3435:10,
                        LowerArmFront:-43.21984:10"
                ));

                AttackMoves.Add(new MoveSet(@"
                        UpperArmFront:286.4763:10,
                        LowerArmFront:0.4966094:10
                    "));
            }

            else if (attackId == AttackIds.thrust)
            {
                TargetEnemy();
                AttackMoves.Add(new MoveSet(@"
                        UpperArmFront:112.2979:20,
                        UpperArm:-0.0164523:20,
                        LowerArm:-0.04736875:20,
                        LowerArmFront:-69.02728:20
                    "));
                AttackMoves.Add(new MoveSet(@"
                        UpperArmFront:-12.23704:20,
                        UpperArm:-0.5381694:20,
                        LowerArm:-4.081913:20,
                        LowerArmFront:-6.531116:20
                    "));
            }

        }


        //
        // ─── ATTACK GO ────────────────────────────────────────────────────────────────
        //
        public void Go()
        {
            frame++;
            if (KB.ActionHeld || KB.Action2Held) power++;

            if (CurrentAttack == AttackIds.club) MeleeClub();
            if (CurrentAttack == AttackIds.thrust) MeleeThrust();
            if (CurrentAttack == AttackIds.dislodgeBack) DislodgeBack();
            if (CurrentAttack == AttackIds.dislodgeKick) DislodgeKick();
        }


        // - - - - - - - - - - - - - - - - - -
        //  MELEE:  CLUB
        // - - - - - - - - - - - - - - - - - -
        public void MeleeClub()
        {
            if (state == AttackState.windup)
            {
                AttackMoves[0].RunMove(Puppet);

                float distance = Mathf.Abs(EnemyTarget.x - Puppet.RB2["UpperBody"].position.x);

                if (!Puppet.IsCrouching && distance < 4.5f && distance > 1.2f)
                {
                    Puppet.RB2["LowerLeg"].AddForce(Vector2.up + (Vector2.right * facing) * Puppet.TotalWeight * distance * 2f);
                    Puppet.RB2["UpperBody"].AddForce(Vector2.up + (Vector2.right * facing) * Puppet.TotalWeight * distance * 2f);
                    Puppet.RB2["Head"].AddForce(Vector2.up * Puppet.TotalWeight * distance * 2f);
                }

                if (!Puppet.IsCrouching && distance < 1.9f)
                {
                    Puppet.RB2["UpperBody"].velocity *= 0.001f;
                }

                if (!Puppet.IsCrouching && distance < 1.5f)
                {
                    Puppet.RB2["LowerLeg"].AddForce(Vector2.up + (Vector2.right * -facing) * Puppet.TotalWeight * distance * 2f);
                    Puppet.RB2["UpperBody"].AddForce(Vector2.up + (Vector2.right * -facing) * Puppet.TotalWeight * distance * 2f);
                    Puppet.RB2["Head"].AddForce(Vector2.up * Puppet.TotalWeight * distance * 2f);
                }


                if (frame == 1)
                {
                    CurrentMove= MoveTypes.slash;
                }



                if (!KB.ActionHeld && !KB.Action2Held)
                {
                    state = AttackState.hit;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.hit)
            {

                if (frame == 1)
                {
                    AttackMoves[0].ClearMove(Puppet);

                    thing.AttackDamage = power;

                    //  Set attack damage power curve
                    if (power > 40) thing.AttackDamage = 40 - (power % 20);

                    Util.Notify("Damage: " + thing.AttackDamage, Util.VerboseLevels.Full);

                    Puppet.RunRigids(BodyInertia);
                }

                if (frame == 2)
                {
                    thing.R.AddForce(((Vector2.right * facing) + (Vector2.down * 0.5f)) * thing.AttackDamage * 2);
                    thing.R.AddTorque(thing.AttackDamage * -facing);

                    Puppet.RB2["LowerArmFront"].AddForce(Vector2.right * facing * Puppet.TotalWeight * 10f);
                    Puppet.RB2["UpperBody"].AddForce(Vector2.right * -facing * Puppet.TotalWeight * 15f);
                    Puppet.RB2["UpperArmFront"].AddTorque(20f * -facing * Puppet.TotalWeight);
                }

                if (frame > 2)
                {
                    float distance = thing.R.position.x - EnemyTarget.x;

                    if (Mathf.Abs(distance) < 2f)
                    {
                        thing.R.mass    = 0.1f;
                        Vector2 target1 = (EnemyTarget - thing.R.position);
                        float force     = 0.01f;// thing.canSlice ? 1.5f : 0.5f;

                        thing.R.AddForceAtPosition(target1 * force, (Vector2)((Vector2)thing.tr.localPosition + (Vector2)(thing.R.transform.right * facing * thing.size)), ForceMode2D.Impulse);

                        Puppet.RB2["LowerArmFront"].AddForce(target1 * thing.AttackDamage * Puppet.TotalWeight);

                        Puppet.RB2["UpperBody"].velocity *= 0.001f;
                    }
                }

                if (frame == 4) thing.R.mass = 0.05f;

                if (frame > 15)
                {
                    state = AttackState.final;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.final)
            {
                thing.AttackDamage = 0f;
                thing.P.MakeWeightless();

                state = AttackState.ready;
                Puppet.RunRigids(BodyInertiaFix);

                Puppet.PBO.OverridePoseIndex = KB.Down ? Puppet.PuppetPose["JT_Duck"] : -1;
            }
        }

        // - - - - - - - - - - - - - - - - - -
        //  MELEE:  THRUST
        // - - - - - - - - - - - - - - - - - -
        public void MeleeThrust()
        {
            if (state == AttackState.windup)
            {
                AttackMoves[0].RunMove(Puppet);


                if (frame == 1)
                {

                    CurrentMove = MoveTypes.stab;
                }

                float distance = Mathf.Abs(EnemyTarget.x - Puppet.RB2["UpperBody"].position.x);

                if (!Puppet.IsCrouching && distance < 3.5f && distance > 1.1f)
                {
                    Puppet.RB2["LowerLeg"].AddForce(Vector2.up + (Vector2.right * facing) * Puppet.TotalWeight * distance * 2f);
                    Puppet.RB2["UpperBody"].AddForce(Vector2.up + (Vector2.right * facing) * Puppet.TotalWeight * distance * 2f);
                    Puppet.RB2["Head"].AddForce(Vector2.up * Puppet.TotalWeight * distance * 2f);
                }
                if (distance < 1.1f)
                {
                    Puppet.RB2["UpperBody"].velocity *= 0.001f;
                }


                if (!KB.ActionHeld && !KB.Action2Held)
                {
                    state = AttackState.hit;
                    frame = 0;
                }


                return;
            }

            if (state == AttackState.hit)
            {
                if (frame == 1)
                {
                    AttackMoves[0].ClearMove(Puppet);

                    thing.AttackDamage = power;

                    if (power > 40) thing.AttackDamage = 40 - (power % 20);

                    Util.Notify("Damage: " + thing.AttackDamage, Util.VerboseLevels.Full);

                    TargetPos    = (EnemyTarget - thing.R.position);
                    TargetAngle  = Vector2.SignedAngle(Vector2.right, TargetPos) - 95f;
                    floatemp1    = Mathf.Abs(EnemyTarget.x - Puppet.RB2["UpperBody"].position.x);
                    thing.R.mass = (floatemp1 > 3f || thing.canStab) ? 0.1f : 0.5f;
                }

                thing.R.MoveRotation(TargetAngle);

                int frameCount = (Enemy == null || floatemp1 > 3f) ? 2 : UnityEngine.Random.Range(6, 8);

                if (frame >= 2 && frame <= frameCount)
                {
                    thing.R.AddForce((TargetPos.normalized + (Vector2.right * facing * 1.5f)) * thing.AttackDamage);
                }
                if (frame > 2)
                {
                    Puppet.RB2["LowerLeg"].velocity      *= 0f;
                    Puppet.RB2["LowerLegFront"].velocity *= 0f;
                }

                if (frame > 30)
                {
                    state = AttackState.final;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.final)
            {
                thing.AttackDamage = 0f;
                thing.P.MakeWeightless();

                state = AttackState.ready;
                Puppet.RunRigids(BodyInertiaFix);

                Puppet.PBO.OverridePoseIndex = KB.Down ? Puppet.PuppetPose["JT_Duck"] : 0;
            }
        }


        // - - - - - - - - - - - - - - - - - -
        //  DISLODGE:  BACK
        // - - - - - - - - - - - - - - - - - -
        public void DislodgeBack()
        {
            if (state == AttackState.windup)
            {

                if (frame == 1)
                {
                    CurrentMove = MoveTypes.stab;
                }

                if ((!KB.ActionHeld && !KB.Action2Held) || power > 40)
                {
                    state = AttackState.hit;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.hit)
            {

                if (frame == 1)
                {
                    if (power < 20) power = 20;
                    thing.AttackDamage = power;

                    //  Set attack damage power curve
                    if (power > 40) thing.AttackDamage = 30 - (power % 40);
                }

                if (frame >= 2)
                {
                    thing.R.AddForce(Vector2.right * -facing * thing.AttackDamage * 2 * Puppet.TotalWeight);
                    
                    thing.CheckLodged();
                    if (!thing.isLodged) state = AttackState.final;
                }

                if (frame > 15)
                {
                    state = AttackState.final;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.final)
            {
                thing.AttackDamage = 0f;
                thing.P.MakeWeightless();

                state = AttackState.ready;

            }
        }


        // - - - - - - - - - - - - - - - - - -
        //  DISLODGE:  Kick
        // - - - - - - - - - - - - - - - - - -
        public void DislodgeKick()
        {
            if (state == AttackState.windup)
            {
                CurrentMove = MoveTypes.stab;

                if ((!KB.ActionHeld && !KB.Action2Held) || power > 40)
                {
                    state = AttackState.hit;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.hit)
            {

                if (frame == 1)
                {
                    TargetPos = (thing.StabVictim.rigidbody.position - Puppet.RB2["FootFront"].position);

                    if (power < 20) power = 20;

                    thing.AttackDamage = power;

                    if (power > 40) thing.AttackDamage = 30 - (power % 40);
                }

                Puppet.LB["UpperLegFront"].Broken = true;

                if (frame >= 2 && frame <= 7) Puppet.RB2["UpperLegFront"].AddTorque(30f * facing);
                

                if (frame == 10)
                {
                    if (thing.StabVictim != null)
                    {
                        thing.StabVictim.rigidbody.AddForce(Vector2.right * facing * thing.AttackDamage * 20);
                        Puppet.RB2["LowerBody"].AddForce(Vector2.right * -facing * thing.AttackDamage * 10);
                    }
                }

                if (frame > 15)
                {
                    state = AttackState.final;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.final)
            {
                thing.AttackDamage = 0f;
                thing.P.MakeWeightless();
                
                Puppet.LB["UpperLegFront"].Broken          = false;
                Puppet.LB["UpperLegFront"].Joint.useLimits = true;
                Puppet.LB["LowerLegFront"].Broken          = false;
                Puppet.LB["LowerLegFront"].Joint.useLimits = true;

                state = AttackState.ready;
            }
        }

        public void BodyInertia(Rigidbody2D rb)
        {
            string exclude = "LowerArmFront UpperArmFront";
            if (exclude.Contains(rb.name)) return;
            rb.inertia = 0.5f;
        }

        public void BodyInertiaFix(Rigidbody2D rigid)
        {
            Puppet.SetRigidOriginal(rigid.name);
        }


        //
        // ─── MOVESET CLASS ────────────────────────────────────────────────────────────────
        //
        public class MoveSet
        {
            public float neverEnding                  = 0;
            public float lastOffset                   = 0f;
            public int framesOK                       = 0;
            private readonly List<LimbMove> moves     = new List<LimbMove>();
            public struct LimbMove
            {
                public string limbName;
                public float angle;
                public float force;
            }

            public MoveSet(string movesList)
            {
                Import(movesList);
            }

            public void Import(string movesList)
            {
                string[] moveItem = movesList.Trim().Split(',');
                foreach (string item in moveItem)
                {
                    if (item.Trim() == "") continue;

                    string[] parts = item.Trim().Split(':');
                    LimbMove move  = new LimbMove()
                    {
                        limbName = parts[0]
                    };
                    if (float.TryParse(parts[1], out float angle)) move.angle = angle;
                    if (float.TryParse(parts[2], out float force)) move.force = force;

                    moves.Add(move);
                }
            }

            public bool RunMove(Puppet Puppet)
            {
                JointMotor2D motor;
                HingeJoint2D joint;
                JointAngleLimits2D limits;

                for (int i = moves.Count; --i >= 0;)
                {
                    joint  = Puppet.LB[moves[i].limbName].Joint;
                    limits = joint.limits;

                    limits.min      = limits.max = moves[i].angle * -Puppet.Facing;
                    joint.limits    = limits;
                    joint.useLimits = false;
                    joint.useMotor  = false;

                    Puppet.LB[moves[i].limbName].Joint = joint;

                    Puppet.LB[moves[i].limbName].InfluenceMotorSpeed(Mathf.DeltaAngle(
                        Puppet.LB[moves[i].limbName].Joint.jointAngle, moves[i].angle * Puppet.PBO.AngleOffset) * moves[i].force);

                    //motor            = Puppet.LB[moves[i].limbName].Joint.motor;
                    //motor.motorSpeed = Mathf.Lerp(motor.motorSpeed, moves[i].angle, moves[i].force);

                    //Puppet.LB[moves[i].limbName].Joint.motor = motor;
                }

                return true;

            }

            public void ClearMove(Puppet Puppet)
            {
                JointMotor2D motor;
                HingeJoint2D joint;
                JointAngleLimits2D limits;

                for (int i = moves.Count; --i >= 0;)
                {
                    joint  = Puppet.LB[moves[i].limbName].Joint;
                    limits = joint.limits;

                    limits.min      = limits.max = 0f;
                    joint.limits    = limits;
                    joint.useLimits = false;
                    joint.useMotor  = false;

                    Puppet.LB[moves[i].limbName].Joint = joint;


                }

                Puppet.PBO.OverridePoseIndex = -1;
            }
        }


    }

}
